"""
Implements various DNS checks with potential DNSSEC
validation, bypassing local cache for the cases where
one needs immediate revalidation or per DNS configurations.

Needs DNSSEC-aware resolver to have a chance to grab domain
cached trust anchors and verify keys.

For now this requires that a cached DNSKEY is still
signing domain data.

Optional support and dependency on PySocks for socks5
proxy support.
"""

from __future__ import absolute_import

import contextlib

import dns.dnssec
import dns.exception
import dns.flags
import dns.message
import dns.name
import dns.query
import dns.rcode
import dns.rdataclass
import dns.rdatatype
import dns.resolver

from . import dnssec
from . import exceptions
from . import monkeypatch  # noqa
from . import resolver
from .resolver import set_nameservers, set_socks5_server, query  # noqa


__version__ = '0.8'


@contextlib.contextmanager
def as_dnssec(c):
    yield Checker(c.domain, dnssec=True, direct=c.direct,
                  errors=c.err_fn, nameservers=c.nameservers)


class Checker(object):
    def __init__(self, domain, dnssec=False, direct=True,
                 errors=None, nameservers=None):
        self.domain = domain
        self.dnssec = dnssec
        self.direct = direct
        self.err_fn = errors
        self.nameservers = nameservers
        self._ns_addrs = None
        self._ns = None

    @property
    def ns(self):
        if not self._ns:
            self._ns = resolver.ns_for(self.domain, self.dnssec)
        return self._ns

    @property
    def ns_addrs(self):
        if not self._ns_addrs:
            self._ns_addrs = self.nameservers \
                or resolver.ns_addrs_for(self.domain, self.dnssec)
        return self._ns_addrs

    def set_nameservers(self, ns):
        self.ns_addrs = ns

    def query_at(self, qname, rdtype, nameserver, timeout=2):
        """Lookup, but explicitely sends a packet to the selected
        nameserver."""
        with resolver.Resolver(timeout) as r:
            ans = r.query_at(qname, rdtype, nameserver, self.dnssec)
        return ans.get()

    def query(self, name, rdtype):
        """Lookup."""
        if self.direct:
            raised = None
            for ns in self.ns_addrs:
                try:
                    return self.query_at(name, rdtype, ns)
                except Exception as e:
                    raised = e
                    pass

            if raised:
                raise raised

        return resolver.query(name, rdtype, self.dnssec)

    def query_relative(self, name, rdtype):
        """Lookup, relative to domain"""
        qname = dns.name.from_text(name, dns.name.from_text(self.domain))
        return self.query(qname, rdtype)

    def notify_error(self, exc):
        if self.err_fn:
            self.err_fn(exc)

    def txt(self, name=''):
        """Return the txt for name under zone, values joined
        as a string, each record separated by a newline"""
        try:
            resp = self.query_relative(name, dns.rdatatype.TXT)
            return '\n'.join(''.join(r.strings) for r in resp
                             if r.rdtype == dns.rdatatype.TXT)
        except (dns.resolver.NoAnswer, exceptions.BadRcode):
            pass

    def has_txt(self, value, also_check=None, ignore_case=True):
        """Find if the specified TXT record in zone exists
        at given names (default @, also_check list of additional
        names). Eventually ignore case."""
        if not also_check:
            also_check = []

        if ignore_case:
            def equals(a, b):
                return a.lower() == b.lower()
        else:
            def equals(a, b):
                return a == b

        for name in [''] + also_check:
            try:
                resp = self.query_relative(name, dns.rdatatype.TXT)
                for rr in resp:
                    if any(equals(value, v) for v in rr.strings):
                        return True
            except (Exception) as e:
                self.notify_error(e)
        return False

    def _uri_to_txt(self, ans):
        txt = ans.rrset[0].target

        # Workaround URI spec change
        return txt[txt.index('http'):]

    def uri(self, name, relative=True):
        """Return the published URI"""
        if relative:
            ans = self.query_relative(name, dns.rdatatype.URI)
        else:
            ans = self.query(name, dns.rdatatype.URI)
        return self._uri_to_txt(ans)

    def mx(self, name=''):
        """Return the mx set for the domain"""
        try:
            mx_set = [rr for rr in self.query_relative(name, 'MX') if
                      rr.rdtype == dns.rdatatype.MX]
        except dns.resolver.NoAnswer:
            return []

        return [rr.exchange.to_text() for rr in
                sorted(mx_set, key=lambda rr: rr.preference)]

    def txt_spf(self):
        """Return first TXT/spf record for domain"""
        for rec in self.txt():
            if rec.startswith('v=spf'):
                return rec

    def spf(self):
        try:
            req = self.query_relative('', 'SPF')
            return req.rrset[0].to_text()
        except dns.resolver.NoAnswer:
            return self.txt_spf()

    def tpda_endpoint(self, name):
        """Return the endpoint according to this domain DNS operator
        setup. (Lookup _tpda service name as an URI on each NS)."""
        answers = []

        with resolver.Resolver(timeout=5) as r:
            for ns in self.ns:
                qname = '{}._tpda._tcp.{}'.format(name, ns)
                if self.direct:
                    answers.append(r.query_at(qname, 'URI',
                                   resolver.ns_for(ns)[0]))
                else:
                    answers.append(r.query(qname, 'URI'))

        for answer in answers:
            try:
                return self._uri_to_txt(answer.get())
            except:
                pass

    def cdnskey(self):
        """ 1. All nameservers should agree on the CDNSKEY set
            2. The CDNSKEY should already be signing the DNSKEY RRSET
            3. Presence of 0 algorithm means remove everything. Cannot
               be mixed with other valid CDNSKEYs.
            4. Older CDNSKEY should not overwrite newer DATA (DS)
               RFC states that. I don't get it.
            5. This should be DNSSEC validated. Yep.
        """
        cds = {}
        dnskeys = {}

        cds_rrset = None

        with resolver.Resolver(timeout=5) as r:
            for ns in self.ns_addrs:
                cds[ns] = r.query_at(self.domain, dns.rdatatype.CDNSKEY,
                                     ns, True)
                dnskeys[ns] = r.query_at(self.domain, dns.rdatatype.DNSKEY,
                                         ns, True)

        for ns in self.ns_addrs:
            # 1.
            cds[ns] = cds[ns].get()
            dnskeys[ns] = dnskeys[ns].get()
            if cds_rrset:
                if set(cds[ns]) != cds_rrset:
                    errstr = ('{} disagrees on CDNSKEY '
                              'RRSet ({}!={})'.format(ns, set(cds), cds_rrset))
                    raise exceptions.BadCDNSKEY(errstr)
            cds_rrset = set(cds[ns])

            # 2.
            for key in cds[ns]:
                if not dnssec.signed_by(dnskeys[ns], key):
                    errstr = ('{} did not sign '
                              'DNSKEY RR'.format(dns.dnssec.key_id(key)))
                    raise exceptions.BadCDNSKEY(errstr)

        for ns in self.ns_addrs:
            # 3.
            if any(key.algorithm == 0 for key in cds[ns]):
                if len(cds[ns]) == 1:
                    # Special case delete
                    raise exceptions.DeleteDS
                else:
                    raise exceptions.BadCDNSKEY('Alg0 and other keys found')

        return cds_rrset



# implements a few zone scanners..
import random

import dns

from . import exceptions
from . import resolver

NAMES = ['@', 'www', 'forum', 'secure', 'mail',
         'git', 'corp', 'localhost', 'autodiscover',
         'test', 'admin', 'support', '_xmpp._tcp',
         '_sip._tcp']


class Scanner:
    types = ['DNAME', 'A', 'AAAA', 'NS', 'MX', 'LOC', 'SPF', 'TXT',
             'URI']

    def __init__(self, domain):
        self.domain = domain
        self.names = NAMES[:]
        self.ns = resolver.ns_for(domain)

    def check_name(self, name, rdtypes):
        answers = []
        with resolver.Resolver() as r:
            qname = dns.name.from_text(name, dns.name.from_text(self.domain))
            for rdtype in rdtypes:
                ns = random.sample(self.ns, 1)[0]
                ans = r.query_at(qname, rdtype, ns)
                answers.append(ans)
        return answers

    def scan(self):
        seen = []
        new = []

        def relname(name):
            dom = dns.name.from_text(self.domain)
            return name.relativize(dom).to_text()

        def refers_to(rrset):
            if rrset.rdtype in (dns.rdatatype.MX,):
                return [rr.exchange for rr in rrset]
            if rrset.rdtype in (dns.rdatatype.CNAME, dns.rdatatype.DNAME,
                                dns.rdatatype.SRV, dns.rdatatype.NS):
                return [rr.target for rr in rrset]
            return []

        def consider(rrset):
            seen.append(relname(rrset.name))
            for name in refers_to(rrset):
                dom = dns.name.from_text(self.domain)
                if name.is_subdomain(dom):
                    if relname(name) not in seen:
                        new.append(relname(name))

        def scan_names(names, types):
            answers = []
            for name in names:
                for answer in self.check_name(name, types):
                    answers.append(answer)

            for answer in answers:
                name = relname(answer.query.q.question[0].name)
                try:
                    rrset = answer.get().rrset
                    consider(rrset)
                    yield name, rrset
                except (exceptions.NoAnswer, exceptions.Timeout) as e:
                    pass
                except exceptions.BadRcode as e:
                    # Bad rcode (refused) for additional name
                    yield name, None

        # Wildcard ?
        for ans in self.check_name('*', ['CNAME']):
            try:
                ans.get()
            except exceptions.NoAnswer as e:
                self.names = ['@', '*']
            except (exceptions.BadRcode):
                pass

        # First, scan for valid names. Remove NXDOMAINS, keep NOANSWER
        # Lookup aliases first so we can consider the right name
        for name, rrset in scan_names(self.names, ['CNAME']):
            if rrset:
                yield rrset
            else:
                self.names.remove(name)

        # Second, in valid names - scan for all possible types
        for name, rrset in scan_names(self.names, self.types):
            if rrset:
                yield rrset

        # Extend with newly discovered names
        for name, rrset in scan_names(new, self.types):
            if rrset:
                yield rrset



import errno
import random
import select
import socket
import struct
import time

import dns
try:
    import socks
    __socks_available = True
except ImportError:
    __socks_available = False

from . import exceptions
from . import dnssec

"""
with dnsknife.resolver.AsyncResolver() as r:
    future = r.query_at('www', 'A', somehost)

future.get()
"""

system_resolver = dns.resolver.Resolver()
pysocks = None


def set_socks5_server(addr, port=1080, username=None, password=None):
    if not __socks_available:
        raise ImportError('No module named socks')

    socks.set_default_proxy(socks.SOCKS5, addr, port, False, username,
                            password)
    global pysocks
    pysocks = socks


can_ipv6 = False
try:
    s = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, 0)
    s.connect(('2::', 53))
    can_ipv6 = True
except:
    pass


def set_nameservers(ns):
    system_resolver.nameservers = ns


def ip_family(addr):
    try:
        socket.inet_pton(socket.AF_INET6, addr)
        return socket.AF_INET6
    except:
        return socket.AF_INET


def make_socket(proto, addr, source=None, source_port=0):
    af = ip_family(addr[0])

    if pysocks:
        sock = pysocks.socksocket(af, proto, 0)
    else:
        sock = socket.socket(af, proto, 0)

    sock.setblocking(0)

    if source is None:
        source = {socket.AF_INET: '0.0.0.0',
                  socket.AF_INET6: '::'}[af]
    try:
        sock.bind((source, source_port,))
        sock.connect(addr)
    except socket.error as e:
        if e.args == errno.EINPROGRESS:
            pass
    return sock


def ns_for(domain, dnssec=False):
    answer = query(domain, dns.rdatatype.NS, dnssec,
                    raise_on_no_answer=False)

    # If we have authority referral, use it
    if (answer.response.rcode() == dns.rcode.NOERROR
        and not answer.rrset):
        return ns_for(answer.response.authority[0].name, dnssec)

    return [ns.target.to_text() for ns in answer.rrset]


def ns_addrs_for(domain, dnssec=False):
    ns_list = ns_for(domain, dnssec)

    addrs = sum((ns_addr_insecure(ns) for ns in ns_list), [])

    if not can_ipv6:
        addrs = filter(lambda addr:
                       ip_family(addr) == socket.AF_INET, addrs)

    return addrs


def query(name, rdtype, dnssec=False, raise_on_no_answer=True):
    """Lookup. Using the locally configured resolvers
    by default. Eventually using the local NS AD bit as a trust source."""
    # Query for our name, let NXDOMAIN raise
    res = system_resolver
    if dnssec:
        res.use_edns(0, dns.flags.DO, 0)

    # Convenience for external callers
    if isinstance(rdtype, str):
        rdtype = dns.rdatatype.from_text(rdtype)

    answer = res.query(name, rdtype, raise_on_no_answer=raise_on_no_answer)

    # Double check we have DNSSEC validation
    if dnssec:
        if not answer.response.flags & dns.flags.AD:
            raise exceptions.NoDNSSEC

    return answer


def ns_addr_insecure(nameserver):
    """Find the nameserver's possible IP addresses. No DNSSEC
    is required here, we'll just validate the end result."""
    ans = []
    for family, socktype, proto, name, sockaddr in \
            socket.getaddrinfo(nameserver, 53):
        if proto == socket.IPPROTO_UDP:
            ans.append(sockaddr[0])
    if len(ans) == 0:
        raise exceptions.NsLookupError(nameserver)
    return ans


class Future:
    def __init__(self, timeout=0):
        if timeout:
            self.deadline = time.time() + timeout
        else:
            self.deadline = None
        self.timed_out = False

    def ready(self):
        if self.deadline and time.time() > self.deadline:
            self.timed_out = True
            return True

        if self._ready():
            fn = getattr(self, 'on_ready', None)
            if fn:
                fn()
                del self.on_ready
            return True
        return False

    def get(self):
        if self.timed_out:
            raise exceptions.Timeout
        return self._get()


class UDPQuery(Future):
    def __init__(self, q, where, source=None, source_port=0, timeout=2):
        Future.__init__(self, timeout)
        self.q = q
        self.rdmsg = b''
        self.net_args = (where, source, source_port)
        self.socket = self.make_socket(*self.net_args)
        self.answer = None
        self.wrmsg = q.to_wire()

    def make_socket(self, *args):
        return make_socket(socket.SOCK_DGRAM, *args)

    def get_sock(self):
        return self.socket

    def retry(self):
        self.socket = self.make_socket(*self.net_args)
        self.wrmsg = self.q.to_wire()

    def writable(self):
        if len(self.wrmsg):
            l = self.socket.send(self.wrmsg)
            self.wrmsg = self.wrmsg[l:]
        return len(self.wrmsg)

    def readable(self):
        buf = self.socket.recv(4096)
        self.rdmsg += buf
        if self.ready():
            return False
        return True

    def _ready(self):
        if len(self.rdmsg):
            try:
                dns.message.from_wire(self.rdmsg)
                return True
            except Exception as e:
                pass
        return False

    def _get(self):
        return dns.message.from_wire(self.rdmsg)


class TCPQuery(UDPQuery):
    def __init__(self, q, where, source, source_port, timeout=10):
        UDPQuery.__init__(self, q, where, source, source_port, timeout)
        self.wrmsg = struct.pack('!H', len(self.wrmsg)) + self.wrmsg
        self.headlen = 0

    def make_socket(self, *args):
        return make_socket(socket.SOCK_STREAM, *args)

    def _ready(self):
        if self.headlen:
            return True
        if len(self.rdmsg) >= 2:
            (l,) = struct.unpack("!H", self.rdmsg[:2])
            if len(self.rdmsg) == l + 2:
                self.headlen = l
                self.rdmsg = self.rdmsg[2:]
                return True
        return False


class FutureAnswer(Future):
    def __init__(self, context, req, where, timeout=0, source=None,
                 source_port=0):
        Future.__init__(self, 0)
        self.context = context
        self.net_args = (where, source, source_port)

        self.req = req
        self.timeout = timeout
        self.attempts = 5

        self.query = UDPQuery(self.req, *self.net_args, timeout=timeout)
        self.query.on_ready = self.notify_ready
        self.context.register(self.query)

    def notify_ready(self):
        try:
            q = self.query.get()
            if q.flags & dns.flags.TC:
                self.query = TCPQuery(self.req, *self.net_args,
                                      timeout=self.timeout)
                self.query.on_ready = self.notify_ready
                self.context.register(self.query)
        except exceptions.Timeout:
            self.attempts -= 1
            if self.attempts > 0:
                self.query.retry()
                self.context.register(self.query)
            else:
                raise

    def _ready(self):
        return self.query.ready()

    def _get(self):
        while not self.ready():
            self.context.one_poll_loop()

        ans = self.query.get()
        if ans.rcode() != dns.rcode.NOERROR:
            raise exceptions.BadRcode(dns.rcode.to_text(ans.rcode()))

        q = self.req
        if not q.is_response(ans):
            raise exceptions.BadAnswer

        answer = dns.resolver.Answer(q.question[0].name, q.question[0].rdtype,
                                     q.question[0].rdclass, ans)

        if self.req.ednsflags & dns.flags.DO:
            if not dnssec.trusted(answer):
                raise exceptions.NoDNSSEC

        return answer


class Resolver:
    def __init__(self, timeout=0, source=None, source_port=0):
        self.futures = {}
        self.poll = select.epoll()
        self.source_port = source_port
        self.source = source
        self.timeout = timeout

    def query_at(self, qname, rdtype, addr, dnssec=False):
        rdclass = dns.rdataclass.IN
        if isinstance(rdtype, str):
            rdtype = dns.rdatatype.from_text(rdtype)
        if isinstance(qname, str):
            qname = dns.name.from_text(qname)
        if isinstance(addr, str):
            if ip_family(addr):
                addr = (addr, 53,)
            else:
                addr = (ns_addr_insecure(addr)[0], 53)

        req = dns.message.make_query(qname, rdtype, rdclass)
        if dnssec:
            req.use_edns(0, dns.flags.DO, 0)

        return FutureAnswer(self, req, addr, self.timeout, self.source,
                            self.source_port)

    def query(self, qname, rdtype, dnssec=False):
        ns = system_resolver.nameservers
        if not len(ns):
            raise exceptions.NsLookupError('no nameservers')
        return self.query_at(qname, rdtype, random.sample(ns, 1)[0], dnssec)

    def register(self, future):
        sock = future.get_sock()
        self.poll.register(sock, select.EPOLLIN | select.EPOLLOUT)
        self.futures[sock.fileno()] = future

    def unregister(self, fd):
        self.poll.unregister(fd)
        del self.futures[fd]

    def one_poll_loop(self):
        if self.futures:
            for (fd, ev) in self.poll.poll(1):
                future = self.futures[fd]
                if ev & select.EPOLLOUT:
                    if not future.writable():
                        self.poll.modify(fd, select.EPOLLIN)
                if ev & select.EPOLLIN:
                    if not future.readable():
                        self.unregister(fd)
                if ev & (select.EPOLLERR | select.POLLHUP):
                    self.unregister(fd)

        for fd, future in self.futures.items():
            if future.ready():
                self.unregister(fd)

    def poll_loop(self):
        while self.futures:
            self.one_poll_loop()

    def __enter__(self):
        return self

    def __exit__(self, exc, exc_val, exc_tb):
        self.poll_loop()



from __future__ import absolute_import

import mock
import os
import unittest

import dns.name
import dns.rdatatype
import dns.rdataclass
import dns.resolver
import dns.rrset

from .. import Checker

here = os.path.dirname(__file__)


def fake_answer(name, rdtype, rvalues):
    name = dns.name.from_text(name)
    rdtype = dns.rdatatype.from_text(rdtype)
    msg = dns.message.make_query(name, rdtype, dns.rdataclass.IN)
    resp = dns.message.make_response(msg)
    rrset = resp.find_rrset(resp.answer, name, dns.rdataclass.IN, rdtype,
                            create=True)
    for rvalue in rvalues:
        rrset.add(dns.rdata.from_text(dns.rdataclass.IN, rdtype, rvalue))
    return dns.resolver.Answer(name, rdtype, dns.rdataclass.IN, resp)


class FakeFuture:
    def __init__(self, name, rdtype, rvalues):
        self.name = name
        self.rdtype = rdtype
        self.rvalues = rvalues

    def get(self):
        return fake_answer(self.name, self.rdtype, self.rvalues)


@mock.patch('dnsknife.resolver.ns_addrs_for', return_value=['1.2.3.4'])
@mock.patch('dnsknife.resolver.Resolver.query_at')
class TestChecker(unittest.TestCase):
    def test_txt(self, rmock, cmock):
        checker = Checker('ten.pm')
        rmock.return_value = FakeFuture('ten.pm', 'TXT', ['pouet'])
        self.assertEqual(checker.txt(), 'pouet')

    def test_query_nodnssec(self, rmock, cmock):
        Checker('ten.pm').query('ten.pm', 'A')
        cmock.assert_called_with('ten.pm', False)
        rmock.assert_called_with('ten.pm', 'A', '1.2.3.4', False)

    def test_query_dnssec(self, rmock, cmock):
        Checker('ten.pm', dnssec=True).query('ten.pm', 'A')
        cmock.assert_called_with('ten.pm', True)
        rmock.assert_called_with('ten.pm', 'A', '1.2.3.4', True)

    def test_query_relative(self, rmock, cmock):
        Checker('ten.pm').query_relative('www', 'A')
        rmock.assert_called_with(dns.name.from_text('www.ten.pm'),
                                 'A', '1.2.3.4', False)

    def test_has_no_txt(self, rmock, cmock):
        rmock.return_value = FakeFuture('ten.pm', 'TXT', ['12345'])
        self.assertFalse(Checker('ten.pm').has_txt('1234', ['www']))

    def test_has_txt(self, rmock, cmock):
        rmock.return_value = FakeFuture('ten.pm', 'TXT', ['1234'])
        self.assertTrue(Checker('ten.pm').has_txt('1234', ['www']))

    def test_has_txt_caseinsensitive(self, rmock, cmock):
        rmock.return_value = FakeFuture('ten.pm', 'TXT', ['1234ABC'])
        self.assertTrue(Checker('ten.pm').has_txt('1234abc', ['www']))

    def test_has_txt_casesensitive(self, rmock, cmock):
        rmock.return_value = FakeFuture('ten.pm', 'TXT', ['1234ABC'])
        self.assertFalse(Checker('ten.pm').has_txt('1234abc', ['www'],
                                                   ignore_case=False))

    def test_mx(self, rmock, cmock):
        rmock.return_value = FakeFuture('ten.pm', 'MX', ['20 a.', '30 c.',
                                                         '10 b.'])
        self.assertEqual(['b.', 'a.', 'c.'], Checker('ten.pm').mx())
